/**
 * Jsonix is a JavaScript library which allows you to convert between XML
 * and JavaScript object structures.
 *
 * Copyright (C) 2010 Aleksei Valikov, Highsource.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.hisrc.jsonix.compiler;

import java.util.HashMap;
import java.util.IdentityHashMap;
import java.util.Map;

import javax.xml.namespace.QName;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.Validate;
import org.hisrc.jscm.codemodel.expression.JSArrayLiteral;
import org.hisrc.jscm.codemodel.expression.JSAssignmentExpression;
import org.hisrc.jscm.codemodel.expression.JSMemberExpression;
import org.hisrc.jscm.codemodel.expression.JSObjectLiteral;
import org.hisrc.jscm.codemodel.statement.JSBlock;
import org.hisrc.jsonix.xjc.customizations.PackageMapping;
import org.jvnet.jaxb2_commons.xml.bind.model.MBuiltinLeafInfo;
import org.jvnet.jaxb2_commons.xml.bind.model.MClassInfo;
import org.jvnet.jaxb2_commons.xml.bind.model.MElementInfo;
import org.jvnet.jaxb2_commons.xml.bind.model.MModelInfo;
import org.jvnet.jaxb2_commons.xml.bind.model.MPackaged;
import org.jvnet.jaxb2_commons.xml.bind.model.MPackagedTypeInfo;
import org.jvnet.jaxb2_commons.xml.bind.model.MPackagedTypeInfoVisitor;
import org.jvnet.jaxb2_commons.xml.bind.model.MPropertyInfo;
import org.jvnet.jaxb2_commons.xml.bind.model.MTypeInfo;
import org.jvnet.jaxb2_commons.xmlschema.XmlSchemaConstants;

public class JsonixCompiler<T, C extends T> {

	final JsonixCode code = new JsonixCode();

	private final MModelInfo<T, C> model;
	private final Map<String, PackageMapping> packageMappings;
	private final Map<String, JsonixModule> modules;
	private final Map<MTypeInfo<T, C>, JSMemberExpression> typeInfoDeclarations = new IdentityHashMap<MTypeInfo<T, C>, JSMemberExpression>();

	private static Map<QName, String> XSD_TYPE_MAPPING = new HashMap<QName, String>();
	{
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.ANYTYPE, "AnyType");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.ANYSIMPLETYPE, "AnySimpleType");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.STRING, "String");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.NORMALIZEDSTRING,
				"NormalizedString");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.TOKEN, "Token");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.LANGUAGE, "Language");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.NAME, "Name");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.NCNAME, "NCName");
		// XSD_TYPE_MAPPING.put(XmlSchemaConstants.ID, "Id");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.ID, "String");
		// XSD_TYPE_MAPPING.put(XmlSchemaConstants.IDREF, "Idref");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.IDREF, "String");
		// XSD_TYPE_MAPPING.put(XmlSchemaConstants.IDREFS, "Idrefs");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.IDREFS, "Strings");
		// XSD_TYPE_MAPPING.put(XmlSchemaConstants.ENTITY, "Entity");
		// XSD_TYPE_MAPPING.put(XmlSchemaConstants.ENTITIES, "Entities");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.NMTOKEN, "NMToken");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.NMTOKENS, "NMTokens");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.BOOLEAN, "Boolean");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.BASE64BINARY, "Base64Binary");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.HEXBINARY, "HexBinary");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.FLOAT, "Float");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.DECIMAL, "Decimal");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.INTEGER, "Integer");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.NONPOSITIVEINTEGER,
				"NonPositiveInteger");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.NEGATIVEINTEGER,
				"NegativeInteger");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.LONG, "Long");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.INT, "Int");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.SHORT, "Short");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.BYTE, "Byte");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.NONNEGATIVEINTEGER,
				"NonNegativeInteger");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.UNSIGNEDLONG, "UnsignedLong");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.UNSIGNEDINT, "UnsignedInt");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.UNSIGNEDSHORT, "UnsignedShort");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.UNSIGNEDBYTE, "UnsignedByte");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.POSITIVEINTEGER,
				"PositiveInteger");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.DOUBLE, "Double");
		// XSD_TYPE_MAPPING.put(XmlSchemaConstants.ANYURI, "AnyURI");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.ANYURI, "String");
		// XSD_TYPE_MAPPING.put(XmlSchemaConstants.QNAME, "QName");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.QNAME, "String");
		// XSD_TYPE_MAPPING.put(XmlSchemaConstants.NOTATION, "Notation");
		// XSD_TYPE_MAPPING.put(XmlSchemaConstants.DURATION, "Duration");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.DURATION, "String");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.DATETIME, "DateTime");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.TIME, "Time");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.DATE, "Date");
		// XSD_TYPE_MAPPING.put(XmlSchemaConstants.GYEARMONTH, "GYearMonth");
		// XSD_TYPE_MAPPING.put(XmlSchemaConstants.GYEAR, "GYear");
		// XSD_TYPE_MAPPING.put(XmlSchemaConstants.GMONTHDAY, "GMonthDay");
		// XSD_TYPE_MAPPING.put(XmlSchemaConstants.GDAY, "GDay");
		// XSD_TYPE_MAPPING.put(XmlSchemaConstants.GMONTH, "GMonth");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.CALENDAR, "Calendar");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.GYEARMONTH, "String");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.GYEAR, "String");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.GMONTHDAY, "String");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.GDAY, "String");
		XSD_TYPE_MAPPING.put(XmlSchemaConstants.GMONTH, "String");
		// XSD_TYPE_MAPPING.put(XmlSchemaConstants.CALENDAR, "String");
	}

	// private

	public JsonixCompiler(final MModelInfo<T, C> model,
			Map<String, PackageMapping> packageMappings) {
		Validate.notNull(model);
		Validate.notNull(packageMappings);
		this.model = model;
		this.packageMappings = new HashMap<String, PackageMapping>(
				packageMappings);
		this.modules = new HashMap<String, JsonixModule>(packageMappings.size());

		for (MBuiltinLeafInfo<T, C> builtinLeafInfo : model
				.getBuiltinLeafInfos()) {
			final String name = XSD_TYPE_MAPPING.get(builtinLeafInfo
					.getTypeName());
			if (name != null) {
				this.typeInfoDeclarations.put(builtinLeafInfo,
						code.$JsonixSchemaXSD.p(name).p("INSTANCE"));
			}
		}
	}

	JsonixModule getModule(final MPackaged packaged) {
		String packageName = packaged.getPackageInfo().getPackageName();
		final boolean blankPackage;
		if (StringUtils.isBlank(packageName)) {
			packageName = "";
			blankPackage = true;
		} else {
			blankPackage = false;
		}
		PackageMapping packageMapping = packageMappings.get(packageName);

		if (packageMapping == null) {

			packageMapping = new PackageMapping();
			packageMapping.setPackageName(packageName);
			packageMappings.put(packageName, packageMapping);
		}

		if (StringUtils.isBlank(packageMapping.getSpaceName())) {
			packageMapping.setSpaceName(blankPackage ? "generated"
					: packageName);
		}

		if (StringUtils.isBlank(packageMapping.getModuleName())) {
			packageMapping.setModuleName("Mappings");
		}
		if (packageMapping.getDirectory() == null) {
			packageMapping.setDirectory((blankPackage ? "" : packageName
					.replace('.', '/')));

		}

		if (packageMapping.getFileName() == null) {
			packageMapping.setFileName(packageMapping.getModuleName() + ".js");

		}
		if (packageMapping.getDeclarationsFileName() == null) {
			packageMapping.setDeclarationsFileName(packageMapping
					.getModuleName() + "-Declarations.js");

		}

		if (packageMapping.getStructuresFileName() == null) {
			packageMapping.setStructuresFileName(packageMapping.getModuleName()
					+ "-Structures.js");
		}

		JsonixModule module = this.modules.get(packageMapping.getSpaceName()
				+ "." + packageMapping.getModuleName());
		if (module == null) {
			module = new JsonixModule(code, packageMapping);
			this.modules.put(module.name, module);
		}
		return module;
	}

	public Map<String, JsonixModule> compile() {

		for (MTypeInfo<T, C> typeInfo : model.getTypeInfos()) {
			typeInfo.acceptTypeInfoVisitor(new MPackagedTypeInfoVisitor<T, C, Void>() {

				@Override
				public Void visitPackagedTypeInfo(MPackagedTypeInfo<T, C> info) {
					final JSMemberExpression typeInfoDeclaration = getTypeInfoDeclaration(info);
					final JsonixModule module = getModule(info);
					// module.typeInfos.append(typeInfoDeclaration);
					return null;
				}

				@Override
				public Void visitClassInfo(MClassInfo<T, C> info) {
					super.visitClassInfo(info);
					final JsonixModule module = getModule(info);
					final JSMemberExpression typeInfoDeclaration = getTypeInfoDeclaration(info);
					final JSBlock classInfoBlock = module.properties.block();
					{
						final MClassInfo<T, C> baseTypeInfo = info
								.getBaseTypeInfo();
						if (baseTypeInfo != null) {
							classInfoBlock.expression(typeInfoDeclaration.p(
									"baseTypeInfo").assign(
									getTypeInfoDeclaration(baseTypeInfo)));
						}
					}
					{
						final JSArrayLiteral properties = code.codeModel
								.array();
						classInfoBlock.expression(typeInfoDeclaration.p(
								"properties").assign(properties));

						for (MPropertyInfo<T, C> propertyInfo : info
								.getProperties()) {
							final JSMemberExpression propertyInfoExpression = propertyInfo
									.acceptPropertyInfoVisitor(new PropertyInfoVisitor<T, C>(
											JsonixCompiler.this,
											JsonixCompiler.this.code,
											getModule(info)));
							properties.append(propertyInfoExpression);
						}
					}
					return null;
				}
			});
		}

		for (MElementInfo<T, C> elementInfo : model.getElementInfos()) {
			JsonixModule module = getModule(elementInfo);
			MTypeInfo<T, C> typeInfo = elementInfo.getTypeInfo();
			MTypeInfo<T, C> scope = elementInfo.getScope();
			QName substitutionHead = elementInfo.getSubstitutionHead();

			final JSObjectLiteral value = code.codeModel.object();
			module.elementInfos.append(value);
			JSAssignmentExpression typeInfoDeclaration = getTypeInfoDeclaration(typeInfo);
			QName elementName = elementInfo.getElementName();
			value.append("elementName",
					module.createJsonixXmlQName(elementName));
			// TODO
			if (typeInfoDeclaration != null) {
				value.append("typeInfo", typeInfoDeclaration);
			}

			if (scope != null) {
				value.append("scope", getTypeInfoDeclaration(scope));
			}
			if (substitutionHead != null) {
				value.append("substitutionHead",
						module.createJsonixXmlQName(substitutionHead));
			}

		}

		return this.modules;

	}

	public JSMemberExpression getTypeInfoDeclaration(MTypeInfo<T, C> typeInfo) {
		JSMemberExpression typeInfoDeclaration = this.typeInfoDeclarations
				.get(typeInfo);
		if (typeInfoDeclaration == null) {
			typeInfoDeclaration = typeInfo
					.acceptTypeInfoVisitor(new CreateTypeInfoDeclarationVisitor<T, C>(
							this, this.code));
			if (typeInfoDeclaration != null) {
				this.typeInfoDeclarations.put(typeInfo, typeInfoDeclaration);
			} else {
				throw new UnsupportedOperationException(
						"Could not find the type declaration for [" + typeInfo
								+ "].");

			}
		}
		return typeInfoDeclaration;

	}
}
