/**
 * Jsonix is a JavaScript library which allows you to convert between XML
 * and JavaScript object structures.
 *
 * Copyright (C) 2010 Aleksei Valikov, Highsource.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.hisrc.jsonix.compiler;

import java.util.Arrays;
import java.util.List;

import javax.xml.namespace.QName;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.Validate;
import org.hisrc.jscm.codemodel.JSCodeModel;
import org.hisrc.jscm.codemodel.JSProgram;
import org.hisrc.jscm.codemodel.expression.JSArrayLiteral;
import org.hisrc.jscm.codemodel.expression.JSAssignmentExpression;
import org.hisrc.jscm.codemodel.expression.JSMemberExpression;
import org.hisrc.jscm.codemodel.expression.JSMemberExpression.MemberNew;
import org.hisrc.jscm.codemodel.expression.JSObjectLiteral;
import org.hisrc.jscm.codemodel.literal.JSStringLiteral;
import org.hisrc.jscm.codemodel.statement.JSStatementGenerator;
import org.hisrc.jsonix.xjc.customizations.PackageMapping;

public class JsonixModule {

	private final JsonixCode code;
	private final JSCodeModel codeModel;
	public final String name;
	public final String packageName;
	public final String spaceName;
	public final String moduleName;
	public final JSProgram declarations;
	public final JSProgram structures;
	public final JSStatementGenerator properties;
	public final JSMemberExpression space;
	public final JSMemberExpression module;
	public final JSArrayLiteral elementInfos;
	// public final JSArrayLiteral typeInfos;
	public final String namespaceURI;
	public final String defaultPrefix;

	public final String directory;
	public final String fileName;
	public final String declarationsFileName;
	public final String structuresFileName;

	public JsonixModule(JsonixCode code, PackageMapping packageMapping) {
		Validate.notNull(code);
		Validate.notEmpty(packageMapping.getSpaceName());
		Validate.notEmpty(packageMapping.getModuleName());
		Validate.notEmpty(packageMapping.getDirectory());
		Validate.notEmpty(packageMapping.getFileName());
		Validate.notEmpty(packageMapping.getDeclarationsFileName());
		Validate.notEmpty(packageMapping.getStructuresFileName());
		this.code = code;
		this.codeModel = code.codeModel;
		this.declarations = codeModel.program();
		this.structures = codeModel.program();

		this.packageName = packageMapping.getPackageName();
		this.spaceName = packageMapping.getSpaceName();
		this.moduleName = packageMapping.getModuleName();
		this.name = spaceName + '.' + moduleName;

		this.directory = packageMapping.getDirectory();
		this.fileName = packageMapping.getFileName();
		this.declarationsFileName = packageMapping.getDeclarationsFileName();
		this.structuresFileName = packageMapping.getStructuresFileName();
		this.space = createSpaceObject();
		this.module = this.space.p(this.moduleName);
		final JSObjectLiteral moduleObject = codeModel.object();
		{
			this.declarations._if(this.module.not())._then().block()
					.expression(this.module.assign(moduleObject));
		}

		this.namespaceURI = packageMapping.getNamespaceURI();
		if (this.namespaceURI != null) {
			moduleObject.append("NAMESPACE_URI",
					codeModel.string(this.namespaceURI));
		}
		this.defaultPrefix = packageMapping.getDefaultPrefix();

		if (this.defaultPrefix != null) {
			moduleObject.append("DEFAULT_PREFIX",
					codeModel.string(this.defaultPrefix));

		}

		// this.typeInfos = codeModel.array();
		this.elementInfos = codeModel.array();
		this.properties = this.structures.block();
		// this.structures
		// .expression(this.module.p("typeInfos").assign(typeInfos));
		this.structures.expression(this.module.p("elementInfos").assign(
				elementInfos));
	}

	private JSMemberExpression createSpaceObject() {
		final List<String> components = Arrays.asList(StringUtils.split(
				spaceName, '.'));
		Validate.notEmpty(components);
		JSMemberExpression current = codeModel
				.globalVariable(components.get(0));
		for (int index = 0; index < components.size(); index++) {
			current = (index == 0 ? current : current.p(components.get(index)));
			this.declarations
					._if(current.typeof().eeq(codeModel.string("undefined")))
					._then().block()
					.expression(current.assign(codeModel.object()));
		}
		return current;
	}

	public MemberNew createJsonixXmlQName(final QName name) {
		Validate.notNull(name);
		final MemberNew value = this.code.$JsonixXMLQName.instantiate();
		final String namespaceURI = name.getNamespaceURI();
		final JSAssignmentExpression $namespaceURI;
		final JSAssignmentExpression $prefix;
		if (namespaceURI.equals(this.namespaceURI)) {
			$namespaceURI = this.module.property("NAMESPACE_URI");
			$prefix = this.module.property("DEFAULT_PREFIX");
		} else {
			$namespaceURI = this.code.codeModel.string(namespaceURI);
			$prefix = this.code.codeModel.string(name.getPrefix());
		}
		final JSStringLiteral $localPart = this.code.codeModel.string(name
				.getLocalPart());
		if (!StringUtils.isEmpty(namespaceURI)
				&& (!StringUtils.isEmpty(name.getPrefix()) ||

				!StringUtils.isEmpty(this.defaultPrefix))) {
			value.args($namespaceURI, $localPart, $prefix);
		} else if (!StringUtils.isEmpty(namespaceURI)) {
			value.args($namespaceURI, $localPart);

		} else {
			value.args($localPart);
		}
		return value;

	}

}
