/**
 * Jsonix is a JavaScript library which allows you to convert between XML
 * and JavaScript object structures.
 *
 * Copyright (C) 2010 Aleksei Valikov, Highsource.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.hisrc.jsonix.compiler;

import javax.xml.namespace.QName;

import org.hisrc.jscm.codemodel.expression.JSArrayLiteral;
import org.hisrc.jscm.codemodel.expression.JSMemberExpression;
import org.hisrc.jscm.codemodel.expression.JSObjectLiteral;
import org.jvnet.jaxb2_commons.xml.bind.model.MAnyAttributePropertyInfo;
import org.jvnet.jaxb2_commons.xml.bind.model.MAnyElementPropertyInfo;
import org.jvnet.jaxb2_commons.xml.bind.model.MAttributePropertyInfo;
import org.jvnet.jaxb2_commons.xml.bind.model.MElementPropertyInfo;
import org.jvnet.jaxb2_commons.xml.bind.model.MElementRefPropertyInfo;
import org.jvnet.jaxb2_commons.xml.bind.model.MElementRefsPropertyInfo;
import org.jvnet.jaxb2_commons.xml.bind.model.MElementTypeInfo;
import org.jvnet.jaxb2_commons.xml.bind.model.MElementTypeInfos;
import org.jvnet.jaxb2_commons.xml.bind.model.MElementsPropertyInfo;
import org.jvnet.jaxb2_commons.xml.bind.model.MMixable;
import org.jvnet.jaxb2_commons.xml.bind.model.MPropertyInfo;
import org.jvnet.jaxb2_commons.xml.bind.model.MPropertyInfoVisitor;
import org.jvnet.jaxb2_commons.xml.bind.model.MTyped;
import org.jvnet.jaxb2_commons.xml.bind.model.MValuePropertyInfo;
import org.jvnet.jaxb2_commons.xml.bind.model.MWildcard;
import org.jvnet.jaxb2_commons.xml.bind.model.MWrappable;

final class PropertyInfoVisitor<T, C extends T> implements
		MPropertyInfoVisitor<T, C, JSMemberExpression> {
	/**
	 * 
	 */
	private final JsonixCompiler<T, C> jsonixCompiler;
	private final JsonixCode code;
	private final JsonixModule module;

	public PropertyInfoVisitor(JsonixCompiler<T, C> jsonixCompiler,
			JsonixCode code, JsonixModule module) {
		this.jsonixCompiler = jsonixCompiler;
		this.code = code;
		this.module = module;
	}

	private void createPropertyInfoOptions(MPropertyInfo<T, C> propertyInfo,
			JSObjectLiteral options) {
		options.append("name",
				this.code.codeModel.string(propertyInfo.getPrivateName()));
		if (propertyInfo.isCollection()) {
			options.append("collection", this.code.codeModel._boolean(true));
		}
	}

	private void createTypedOptions(MTyped<T, C> info, JSObjectLiteral options) {
		options.append("typeInfo",
				this.jsonixCompiler.getTypeInfoDeclaration(info.getTypeInfo()));
	}

	private void createWrappableOptions(MWrappable info, JSObjectLiteral options) {
		final QName wrapperElementName = info.getWrapperElementName();
		if (wrapperElementName != null) {
			options.append("wrapperElementName",
					module.createJsonixXmlQName(wrapperElementName));
		}
	}

	private void createElementTypeInfoOptions(MElementTypeInfo<T, C> info,
			JSObjectLiteral options) {
		final QName elementName = info.getElementName();
		options.append("elementName", module.createJsonixXmlQName(elementName));
		createTypedOptions(info, options);
	}

	private void createWildcardOptions(MWildcard info, JSObjectLiteral options) {
		if (info.isDomAllowed()) {
			options.append("domAllowed", this.code.codeModel._boolean(true));
		}
		if (info.isTypedObjectAllowed()) {
			options.append("typedObjectAllowed",
					this.code.codeModel._boolean(true));
		}
	}

	private void createMixableOptions(MMixable info, JSObjectLiteral options) {
		if (info.isMixed()) {
			options.append("mixed", this.code.codeModel._boolean(true));
		}
	}

	public JSMemberExpression visitElementPropertyInfo(
			MElementPropertyInfo<T, C> info) {
		JSObjectLiteral options = this.code.codeModel.object();
		createPropertyInfoOptions(info, options);
		createWrappableOptions(info, options);
		createElementTypeInfoOptions(info, options);
		return this.code.$JsonixModelElementPropertyInfo.instantiate().args(
				options);
	}

	public JSMemberExpression visitElementsPropertyInfo(
			MElementsPropertyInfo<T, C> info) {
		JSObjectLiteral options = this.code.codeModel.object();
		createPropertyInfoOptions(info, options);
		createWrappableOptions(info, options);

		createElementTypeInfosOptions(info, options);
		return this.code.$JsonixModelElementsPropertyInfo.instantiate().args(
				options);
	}

	private void createElementTypeInfosOptions(MElementTypeInfos<T, C> info,
			JSObjectLiteral options) {
		final JSArrayLiteral elementTypeInfos = this.code.codeModel.array();
		options.append("elementTypeInfos", elementTypeInfos);
		for (MElementTypeInfo<T, C> elementTypeInfo : info
				.getElementTypeInfos()) {
			final JSObjectLiteral elementTypeInfoOptions = this.code.codeModel
					.object();
			createElementTypeInfoOptions(elementTypeInfo,
					elementTypeInfoOptions);
			elementTypeInfos.append(elementTypeInfoOptions);
		}
	}

	public JSMemberExpression visitAnyElementPropertyInfo(
			MAnyElementPropertyInfo<T, C> info) {
		JSObjectLiteral options = this.code.codeModel.object();
		createPropertyInfoOptions(info, options);
		createWildcardOptions(info, options);
		createMixableOptions(info, options);
		return this.code.$JsonixModelAnyElementPropertyInfo.instantiate().args(
				options);
	}

	public JSMemberExpression visitAttributePropertyInfo(
			MAttributePropertyInfo<T, C> info) {
		JSObjectLiteral options = this.code.codeModel.object();
		createPropertyInfoOptions(info, options);
		createTypedOptions(info, options);
		options.append("attributeName",
				module.createJsonixXmlQName(info.getAttributeName()));
		return this.code.$JsonixModelAttributePropertyInfo.instantiate().args(
				options);
	}

	public JSMemberExpression visitAnyAttributePropertyInfo(
			MAnyAttributePropertyInfo<T, C> info) {
		JSObjectLiteral options = this.code.codeModel.object();
		createPropertyInfoOptions(info, options);
		return this.code.$JsonixModelAnyAttributePropertyInfo.instantiate()
				.args(options);
	}

	public JSMemberExpression visitValuePropertyInfo(
			MValuePropertyInfo<T, C> info) {
		JSObjectLiteral options = this.code.codeModel.object();
		createPropertyInfoOptions(info, options);
		createTypedOptions(info, options);
		return this.code.$JsonixModelValuePropertyInfo.instantiate().args(
				options);
	}

	public JSMemberExpression visitElementRefPropertyInfo(
			MElementRefPropertyInfo<T, C> info) {
		JSObjectLiteral options = this.code.codeModel.object();
		createPropertyInfoOptions(info, options);
		createMixableOptions(info, options);
		createWrappableOptions(info, options);
		createWildcardOptions(info, options);
		createElementTypeInfoOptions(info, options);
		return this.code.$JsonixModelElementRefPropertyInfo.instantiate().args(
				options);
	}

	public JSMemberExpression visitElementRefsPropertyInfo(
			MElementRefsPropertyInfo<T, C> info) {
		JSObjectLiteral options = this.code.codeModel.object();
		createPropertyInfoOptions(info, options);
		createMixableOptions(info, options);
		createWrappableOptions(info, options);
		createWildcardOptions(info, options);
		createElementTypeInfosOptions(info, options);
		return this.code.$JsonixModelElementRefsPropertyInfo.instantiate()
				.args(options);
	}
}