/**
 * Jsonix is a JavaScript library which allows you to convert between XML
 * and JavaScript object structures.
 *
 * Copyright (C) 2010 Aleksei Valikov, Highsource.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.hisrc.jsonix.xjc.plugin;

import java.io.IOException;
import java.io.StringWriter;
import java.text.MessageFormat;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.bind.JAXBException;

import org.apache.commons.lang.Validate;
import org.hisrc.jscm.codemodel.JSProgram;
import org.hisrc.jscm.codemodel.writer.CodeWriter;
import org.hisrc.jsonix.compiler.JsonixCompiler;
import org.hisrc.jsonix.compiler.JsonixModule;
import org.hisrc.jsonix.xjc.customizations.JsonixCustomizationsConstants;
import org.hisrc.jsonix.xjc.customizations.PackageMapping;
import org.jvnet.jaxb2_commons.util.CustomizationUtils;
import org.jvnet.jaxb2_commons.xjc.model.concrete.XJCCMInfoFactory;
import org.jvnet.jaxb2_commons.xml.bind.model.MModelInfo;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import com.sun.codemodel.JPackage;
import com.sun.codemodel.fmt.JTextFile;
import com.sun.tools.xjc.BadCommandLineException;
import com.sun.tools.xjc.Options;
import com.sun.tools.xjc.Plugin;
import com.sun.tools.xjc.model.CPluginCustomization;
import com.sun.tools.xjc.model.Model;
import com.sun.tools.xjc.model.nav.NClass;
import com.sun.tools.xjc.model.nav.NType;
import com.sun.tools.xjc.outline.Outline;

public class JsonixPlugin extends Plugin {

	public static final String OPTION_NAME = "Xjsonix";
	public static final String OPTION = "-" + OPTION_NAME;

	@Override
	public String getOptionName() {
		return OPTION_NAME;
	}

	@Override
	public String getUsage() {
		return "TBD";
	}

	@Override
	public void onActivated(Options opts) throws BadCommandLineException {
	}

	@Override
	public void postProcessModel(Model model, ErrorHandler errorHandler) {
	}

	@Override
	public List<String> getCustomizationURIs() {
		return Arrays.asList(JsonixCustomizationsConstants.NAMESPACE_URI);
	}

	@Override
	public boolean isCustomizationTagName(String nsUri, String localName) {
		return (JsonixCustomizationsConstants.NAMESPACE_URI.equals(nsUri) && JsonixCustomizationsConstants.PACKAGE_MAPPING_LOCAL_NAME
				.equals(localName));
	}

	@Override
	public boolean run(Outline outline, Options opt, ErrorHandler errorHandler)
			throws SAXException {

		final Model model = outline.getModel();

		Map<String, PackageMapping> packageMappings = new HashMap<String, PackageMapping>();
		{
			List<CPluginCustomization> customizations = CustomizationUtils
					.findCustomizations(model,
							JsonixCustomizationsConstants.PACKAGE_MAPPING_NAME);

			for (CPluginCustomization customization : customizations) {
				try {
					final PackageMapping packageMapping = JsonixCustomizationsConstants
							.unmarshalPackageMapping(customization.element);
					// TODO check parameters
					packageMappings.put(packageMapping.getPackageName(),
							packageMapping);
				} catch (JAXBException jaxbex) {
					errorHandler.error(new SAXParseException(
							"Could not unmarshal the customization element.",
							customization.locator, jaxbex));
				}
			}
		}

		MModelInfo<NType, NClass> mModel = new XJCCMInfoFactory(model)
				.createModel();

		final JsonixCompiler<NType, NClass> compiler = new JsonixCompiler<NType, NClass>(
				mModel, packageMappings);
		final Map<String, JsonixModule> modules = compiler.compile();

		for (JsonixModule module : modules.values()) {
			try {
				final JPackage _package = model.codeModel
						._package(module.packageName);
				_package.addResourceFile(createTextFile(module.fileName,
						module.declarations, module.structures));
				_package.addResourceFile(createTextFile(
						module.declarationsFileName, module.declarations));
				_package.addResourceFile(createTextFile(
						module.structuresFileName, module.structures));
			} catch (IOException ioex) {
				errorHandler.error(new SAXParseException(MessageFormat.format(
						"Could not create the code for the module [{0}].",
						module.name), null, ioex));
			}

		}

		return true;
	}

	private JTextFile createTextFile(String fileName, JSProgram... programs)
			throws IOException {
		Validate.notNull(fileName);
		final JTextFile textFile = new JTextFile(fileName);
		final StringWriter stringWriter = new StringWriter();
		final CodeWriter codeWriter = new CodeWriter(stringWriter);
		for (JSProgram program : programs) {
			codeWriter.program(program);
			codeWriter.lineTerminator();
		}
		textFile.setContents(stringWriter.toString());
		return textFile;
	}

}
